/* -----------------------------------------

   Create deltaC x-tables for all available values of alphas

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h> 
#include <getopt.h> 

#include "hell.hh"


using namespace std;


string sas(double as) {
  ostringstream os;
  int ias = int(1000*as);
  if(1000*as-ias>0.5) ias += 1;
  if     (as<0.01) os << "000" << ias;
  else if(as<0.1 ) os << "00"  << ias;
  else if(as<1.  ) os << "0"   << ias;
  else os << ias;
  return os.str();
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -m  --massive          compute massive C.F.s" << endl
       << " -a  --alphas <as>      set alpha_s" << endl
       << " -n  --nf <nf>          set the number of active flavours" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], bool &massive, double &as, int &nf) {
  const char* const short_options = "hma:n:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "massive", 0, NULL, 'm' },
					 { "alphas", 1, NULL, 'a' },
					 { "nf", 1, NULL, 'n' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'm':
      massive = true;
      break;
    case 'a':
      as = strtod(optarg, NULL);
      break;
    case 'n':
      nf = int(strtod(optarg, NULL));
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  return;
}



int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  bool massive = false;
  int nf0 = -1;
  double as0 = -1;
  read_arguments(argc, argv, massive, as0, nf0);

  HELL::LogOrder ord = HELL::NLL;


  // x grid
  int Np1 = 300, Np2 = 20-1;
  double x;
  double x_min = 1e-9;
  double x_mid = 0.1;
  double x_max = 0.99+0.01;

  for(int nf=3; nf<=6; nf++) {
    if(nf0>0 && nf!=nf0) continue;
    HELL::HELLnf sxD(nf, ord, "./data/");
    sxD.SetLargexDamping();
    //sxD.SetMomentumConservation();
    vector<double> as1, as2;
    sxD.GetAvailableAlphas(as1, as2);
    // alphas grid
    double as_min = as1[0];
    double as_max = as1[as1.size()-1];
    if(as2[0]>as_min) as_min = as2[0];
    if(as2[as2.size()-1]<as_max) as_max = as2[as2.size()-1];
    int Nas = (as_max-as_min)*1000 +1;
    Nas = (as_max-as_min)*100 +1; // smaller set; comment out for a (10 times) larger set
    //
    for(int a=0; a<Nas; a++) {
      double as = as_min + a*(as_max-as_min)/(Nas-1.);
      if(as0>0 && as!=as0) continue;
      //as = int(1000*as)/1000.;  // useless now
      //cout << setprecision(10) << 1000*as << "  " << sas(as) << endl;
      ostringstream filename;
      filename << "./data/x-space/xtableC_nlo_nf" << nf << "_alphas" << sas(as) << ".table";
      ofstream ofile(filename.str());
      ofile
	<< setw(15) << Np1
	<< setw(15) << Np2
	<< setw(15) << x_min
	<< setw(15) << x_mid
	<< setw(15) << x_max
	<< endl;
      //
      for(int i=0; i<Np1+Np2; i++) {
	if(i<Np1) x = x_min * exp(i/(Np1-1.)*log(x_mid/x_min));
	else      x = x_mid + (i-Np1+1)*(x_max-x_mid)/(Np2-0.);
	//cout << x << endl;
	double dC2g = x*sxD.deltaC2g(as, x);
	double dCLg = x*sxD.deltaCLg(as, x);
	ofile
	  //<< setw(15) << x
	  << setw(15) << dC2g
	  << setw(15) << dCLg;
	ofile << endl;
      }
      cout << ".";
      cout.flush();
      //
      ofile.close();
    }
    cout << endl;
  }


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
