#include <iostream>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_integration.h>

#include "math/borel.hh"
#include "integration.hh"
//#include "function_calculus.hh"
//#include "special_functions.hh"
//#include "struct_definitions.hh"
#include "math/pade_approximant.hh"


using namespace std;



namespace HELL {

  unsigned int ORDER = 1;

  void SetBorelOrder(unsigned int order) {
    if(order < 1 || order > 2) {
      cout << "ERROR: Borel order must be 1 or 2." << endl;
      exit(909);
    }
    ORDER = order;
  }

  double bk(int k) {
    int n = ORDER;
    double damp = exp( - (double)n * gsl_sf_lnfact(k) );
    return damp;
  }


  // Borel kernel B_n(w)
  double kernelBn(double w) {
    double Bn = 0;
    if(ORDER == 1) {
      Bn = exp(-w);
    } else if(ORDER == 2) {
      Bn = 2.*gsl_sf_bessel_K0(2.*sqrt(w));
    }
    return Bn;
  }







  // Borel-Pade summation
  struct s_BorelPade_pars {
    int nPadeNum, nPadeDen;
    int Ncoeff;
    dcomplex *coeff;
    dcomplex z;
    bool *fail;
    dcomplex *a;
    dcomplex *b;
  };
  dcomplex BorelPade_integrand(double w, void *p) {
    s_BorelPade_pars btt = *(s_BorelPade_pars *) p;
    return kernelBn(w) * BorelPadeTransform(w*btt.z, btt.nPadeNum, btt.nPadeDen, btt.coeff, btt.Ncoeff, btt.a, btt.b, btt.fail);
  }
  double BorelPade_integrand_re(double w, void *p) {
    return real(BorelPade_integrand(w,p));
  }
  double BorelPade_integrand_im(double w, void *p) {
    return imag(BorelPade_integrand(w,p));
  }
  //pthread_mutex_t mutex_borel;
  dcomplex BorelPadeSum(dcomplex z, int nPade, dcomplex *coeff, int Ncoeff) {
    return BorelPadeSum(z, nPade, nPade, coeff, Ncoeff);
    /*
    if(Ncoeff < 2*nPade+1) {
      cout << "ERROR: Borel-Pade': number of coefficients Ncoeff is too small compared to the Pade' order." << endl;
      exit(566);
    }
    dcomplex *transf_coeff = new dcomplex[2*nPade+1];
    for(int k=0; k<=2*nPade; k++) {
      transf_coeff[k] = coeff[k] * bk(k);
    }
    // linear system
    dcomplex *a = new dcomplex[nPade+1];
    dcomplex *b = new dcomplex[nPade];
    //
    //       GiNaC (mutex mandatory)
    //       Real time elapsed:  93.2276 s
    //       CPU  time elapsed:  144.268 s
    //       ---
    //       gsl (with or without mutex)
    //       Real time elapsed:  41.9392 s
    //       CPU  time elapsed:  101.857 s
    //
    //pthread_mutex_lock(&mutex_borel);
    PadeCoefficients(transf_coeff, nPade, a, b);
    //pthread_mutex_unlock(&mutex_borel);
    //
    bool fail = 0;
    s_BorelPade_pars btt = { nPade, nPade, Ncoeff, transf_coeff, z, &fail, a, b };
    double error;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double res_re = integration_gauss_inf(BorelPade_integrand_re, 0., 1.e-4, &error, &btt);
    double res_im = integration_gauss_inf(BorelPade_integrand_im, 0., 1.e-4, &error, &btt);
    //for(int i=0; i< 100; i++) {
    //  double wz = 0.9 + i/301.;
    //  cout << wz << "   \t" << real(BorelPadeTransform(wz, nPade, transf_coeff, Ncoeff, a, b, &fail)) << endl;
    //}
    gsl_set_error_handler (old_handler);
    if(fail) cout << "WARNING: Possible loss of precision due to small denominator in PadeEpsilonAlgorithm" << endl;
    return dcomplex(res_re, res_im);
    */
  }
  dcomplex BorelPadeTransform(dcomplex wz, int nPadeNum, int nPadeDen, dcomplex *coeff, int Ncoeff, dcomplex *a, dcomplex *b, bool *fail) {
    if(Ncoeff < nPadeNum+nPadeDen+1) {
      cout << "ERROR: Borel-Pade': number of coefficients Ncoeff is too small compared to the Pade' order." << endl;
      exit(566);
    }
    dcomplex res = 0;
    // epsilon algorithm
    //res = PadeEpsilonAlgorithm<dcomplex,dcomplex>(wz, coeff, 2*nPadeNum, fail);
    //*
    // linear system
    res = PadeApproximant(wz, a, b, nPadeNum, nPadeDen);
    //*/
    if(*fail) cout << "WARNING: Possible loss of precision due to small denominator in PadeEpsilonAlgorithm" << endl;
    return res;
  }


  dcomplex BorelPadeSum(dcomplex z, int nPadeNum, int nPadeDen, dcomplex *coeff, int Ncoeff) {
    if(Ncoeff < nPadeNum+nPadeDen+1) {
      cout << "ERROR: Borel-Pade': number of coefficients Ncoeff is too small compared to the Pade' order." << endl;
      exit(566);
    }
    dcomplex *transf_coeff = new dcomplex[nPadeNum+nPadeDen+1];
    for(int k=0; k<=nPadeNum+nPadeDen; k++) {
      transf_coeff[k] = coeff[k] * bk(k);
    }
    // linear system
    dcomplex *a = new dcomplex[nPadeNum+1];
    dcomplex *b = new dcomplex[nPadeDen];
    //
    //pthread_mutex_lock(&mutex_borel);
    PadeCoefficients(transf_coeff, nPadeNum, nPadeDen, a, b);
    //pthread_mutex_unlock(&mutex_borel);
    //
    bool fail = 0;
    s_BorelPade_pars btt = { nPadeNum, nPadeDen, Ncoeff, transf_coeff, z, &fail, a, b };
    double error;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1.e-3; //1.e-4;
    double res_re = integration_gauss_inf(BorelPade_integrand_re, 0., prec, &error, &btt);
    double res_im = integration_gauss_inf(BorelPade_integrand_im, 0., prec, &error, &btt);
    //for(int i=0; i< 100; i++) {
    //  double wz = 0.9 + i/301.;
    //  cout << wz << "   \t" << real(BorelPadeTransform(wz, nPade, transf_coeff, Ncoeff, a, b, &fail)) << endl;
    //}
    gsl_set_error_handler (old_handler);
    if(fail) cout << "WARNING: Possible loss of precision due to small denominator in PadeEpsilonAlgorithm" << endl;
    return dcomplex(res_re, res_im);
  }



};
